#include <stddef.h>
#include <stdio.h>

#include <thread>
#include <chrono>
#include <fstream>
#include <sstream>

#include <cmath>

#include "pid_controller.hpp"

#include "TRPCLongWrapper.hpp"
#include "TRPCLong_GeneralWrapper.h"
#include "dpc_trpc_long/executable.h"

#ifdef __cplusplus
extern "C"
{
#endif

// C header here
#include "TRPC_Long_Itg.h"

#ifdef __cplusplus
}
#endif

typedef enum {
 IPC_Target_Motor_En,       // uint8        --> AccVlcMainReq = 1 
 IPC_Target_Gear,           // uint8        --> AccVlcMainReq = 1 -> D, else -> N
 IPC_Target_Mode,           // uint8        --> Set as 1
 IPC_Life_Signal,           // uint8        --> Counter 0-15
 IPC_Target_Speed,          // float32      --> Vx_des_tar
 IPC_Target_SpdValid,       // uint8        --> Set as 1
 IPC_Target_Brk_Press,      // double       --> Ax_tar < 0 and VlcDCEnable == 1
 IPC_Target_Angle,          // float32      --> EPSRoadWheelReqValue
 IPC_Target_Steering_En,    // uint8        --> EPSRoadWheelReqSts
 IPC_Target_Torque,         // double       --> MProp
 IPC_Torque_Rpm_Mode,       // uint8        --> Set 1  
 IPC_TurnLight,             // uint8        --> Brake pressure and DC enable
 IPC_Target_SpdMode         // uint8        --> based on Vx_des_tar value range
} PendingSignals;

struct TRPCLongSWC2CANStruct {
    SWC2CAN ControlMode;
    uint8_t control_mode;

    SWC2CAN TargetGear;
    uint8_t target_gear_value;

    SWC2CAN DriveMotorEnable;
    uint8_t drive_motor_enable;

    SWC2CAN TorqueOrRPMMode;
    uint8_t torque_or_rpm_mode;

    SWC2CAN SpeedMode;
    uint8_t speed_mode;

    SWC2CAN TargerSpeed;
    float Vx_des_tar;

    SWC2CAN TargerSpeedValid;
    uint8_t TargerSpeedValid_value;

    SWC2CAN TargetTorque;
    double MProp;

    SWC2CAN BrakePressure;
    double brake_pressure_command;

    SWC2CAN TurnLight;
    uint8_t turn_signal;

    SWC2CAN LifeCycle;
    uint8_t life_cycle;
};

std::string LogFileNameDated(const std::string& inputName);
void logTRPCLongData(std::ofstream *ptr_File, uint8_t flag_First, dpc::VED4DPC* const ptr_trpclong_data_from_VED,
                    dpc::Decision4Long* const  ptr_trpclong_data_from_DM,
                    dpc::DM2LongRefLine* const  trpclong_data_from_DM_refLine,
                    dpc::TargetsInfo* const ptr_trpclong_data_from_EM_TargetsInfo, 
                    dpc::TargetInfo* const ptr_trpclong_data_from_EM_TargetInfo, 
                    dpc::RoadEdgeInfo* const ptr_trpclong_data_from_Em_RoadEdgeInfo, 
                    dpc::LaneMarkerInfo* const ptr_trpclong_data_from_EM_LaneMarkerInf,
                    TRPCLongSWC2CANStruct* const ptr_trpclong_data_to_CAN
					);
uint64_t getCurrentTimeStamp();
void assignTRPCLongInputsFromBAandEM(const dpc::Decision4Long data_From_DM_4_TRPCLong,
                                     const dpc::TargetsInfo data_From_EMTargetsInfo_4_TRPCLong,
                                     const dpc::TargetInfo data_From_EMTargetInfo_4_TRPCLong,
                                     const dpc::RoadEdgeInfo data_From_EMRoadEdgeInfo_4_TRPCLong, 
                                     const dpc::LaneMarkerInfo data_Form_EMLaneMarkerInfo_4_TRPCLong) {                                                          
  TRPC_Long_Itg_U.ba_trp_v_sv_setspd_in = data_From_DM_4_TRPCLong.ba_trp_v_sv_setspd;      // ba_trp_v_sv_setspd = 1.4, final op: 3.6, set speed dm: 5
  TRPC_Long_Itg_U.ba_trp_indx_trp_lgt_sts_in = data_From_DM_4_TRPCLong.ba_trp_indx_trp_lgt_sts;
  TRPC_Long_Itg_U.ba_trp_a_tgt_longt_accel_iv_in = data_From_DM_4_TRPCLong.ba_trp_a_tgt_longt_accel_iv;                              
  TRPC_Long_Itg_U.ba_trp_indx_tgt_id_iv_in = data_From_DM_4_TRPCLong.ba_trp_indx_tgt_id_iv;
  TRPC_Long_Itg_U.ba_trp_v_tgt_longt_spd_iv_in = data_From_DM_4_TRPCLong.ba_trp_v_tgt_longt_spd_iv;
  TRPC_Long_Itg_U.ba_trp_d_tgt_longt_dist_iv_in = data_From_DM_4_TRPCLong.ba_trp_d_tgt_longt_dist_iv;
  TRPC_Long_Itg_U.ba_trp_indx_spdlim_typ_in = data_From_DM_4_TRPCLong.ba_trp_indx_spdlim_typ;
  TRPC_Long_Itg_U.ba_trp_mps_spdlim_act_val_in = data_From_DM_4_TRPCLong.ba_trp_kmh_spdlim_act_val/3.6;
  TRPC_Long_Itg_U.ba_trp_d_spdlim_nxt_dst_in = data_From_DM_4_TRPCLong.ba_trp_d_spdlim_nxt_dst;
  TRPC_Long_Itg_U.ba_trp_indx_opt_action_in = data_From_DM_4_TRPCLong.ba_trp_indx_opt_action;

  TRPC_Long_Itg_U.rm_iv_indx_tgt_id = data_From_EMTargetInfo_4_TRPCLong.indx_tgtid; // Assumed the assignment from TargetInfo is OK
  TRPC_Long_Itg_U.rm_iv_d_tgtLongtdist = data_From_EMTargetInfo_4_TRPCLong.d_tgtlongtdist;
  TRPC_Long_Itg_U.rm_iv_v_tgtlongtspd = data_From_EMTargetInfo_4_TRPCLong.v_tgtlongtspd;
  TRPC_Long_Itg_U.rm_iv_a_tgtlongtaccel = data_From_EMTargetInfo_4_TRPCLong.a_tgtlongtaccel;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_v_sv_setspd_in: " << +TRPC_Long_Itg_U.ba_trp_v_sv_setspd_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_indx_trp_lgt_sts_in:" << +TRPC_Long_Itg_U.ba_trp_indx_trp_lgt_sts_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_a_tgt_longt_accel_iv_in: " << +TRPC_Long_Itg_U.ba_trp_a_tgt_longt_accel_iv_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_indx_tgt_id_iv_in: " << +TRPC_Long_Itg_U.ba_trp_indx_tgt_id_iv_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_v_tgt_longt_spd_iv_in: " << +TRPC_Long_Itg_U.ba_trp_v_tgt_longt_spd_iv_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_d_tgt_longt_dist_iv_in: " << +TRPC_Long_Itg_U.ba_trp_d_tgt_longt_dist_iv_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_indx_spdlim_typ_in: " << +TRPC_Long_Itg_U.ba_trp_indx_spdlim_typ_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_mps_spdlim_act_val_in: " << +TRPC_Long_Itg_U.ba_trp_mps_spdlim_act_val_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_d_spdlim_nxt_dst_in: " << +TRPC_Long_Itg_U.ba_trp_d_spdlim_nxt_dst_in<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ba_trp_indx_opt_action_in:" << +TRPC_Long_Itg_U.ba_trp_indx_opt_action_in << std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.rm_iv_indx_tgt_id: " << +TRPC_Long_Itg_U.rm_iv_indx_tgt_id<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.rm_iv_d_tgtLongtdist: " << +TRPC_Long_Itg_U.rm_iv_d_tgtLongtdist<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.rm_iv_v_tgtlongtspd: " << +TRPC_Long_Itg_U.rm_iv_v_tgtlongtspd<< std::endl;
  std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.rm_iv_a_tgtlongtaccel: " << +TRPC_Long_Itg_U.rm_iv_a_tgtlongtaccel<< std::endl;
}

void assignTRPCLongInputsFromVED(const dpc::VED4DPC data_FromVED4TRPCLong) {
    uint8_t _is_standstill = 0;
    if (data_FromVED4TRPCLong.MotionState == 3  && std::fabs(data_FromVED4TRPCLong.LongitudinalVelocity)<0.05) //check whether 3 is correct or not
    {
        _is_standstill = 1;
    }
    TRPC_Long_Itg_U.ivehsdslsts_in = _is_standstill; 
    TRPC_Long_Itg_U.v_ego_mps = data_FromVED4TRPCLong.LongitudinalVelocity;  
    TRPC_Long_Itg_U.a_ego = data_FromVED4TRPCLong.LongitudinalAcceleration;
    TRPC_Long_Itg_U.AxSlopeDyn = 0.; // If required, we can calculate it using IMU data
    TRPC_Long_Itg_U.SpdActl_Esti_mps_m = data_FromVED4TRPCLong.LongitudinalVelocity;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.ivehsdslsts_in: " << +TRPC_Long_Itg_U.ivehsdslsts_in<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.v_ego_mps: " << +TRPC_Long_Itg_U.v_ego_mps<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.a_ego: " << +TRPC_Long_Itg_U.a_ego<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.AxSlopeDyn: " << +TRPC_Long_Itg_U.AxSlopeDyn<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.SpdActl_Esti_mps_e: " << +TRPC_Long_Itg_U.SpdActl_Esti_mps_m<< std::endl;
}

void assignTRPCLongInputsFromCAN(const CAN2SWC canData) {

    TRPC_Long_Itg_U.istrgwhlang_in = canData.VCU_Info_1_Signals.current_angle;   
    TRPC_Long_Itg_U.AccelPedal = 0.;
    TRPC_Long_Itg_U.iwhlbrkprssts = 2;
    TRPC_Long_Itg_U.MPropAct = canData.VCU_Info_1_Signals.motor_torque;
    TRPC_Long_Itg_U.MPropMin = 0.;
    TRPC_Long_Itg_U.MPropMax = 12.5; //1000.;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.istrgwhlang_in: " << +TRPC_Long_Itg_U.istrgwhlang_in<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.AccelPedal: " << +TRPC_Long_Itg_U.AccelPedal<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.iwhlbrkprssts: " << +TRPC_Long_Itg_U.iwhlbrkprssts<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.MPropAct: " << +TRPC_Long_Itg_U.MPropAct<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.MPropMin: " << +TRPC_Long_Itg_U.MPropMin<< std::endl;
    std::cout << "@@@@ LONG SDK IN: TRPC_Long_Itg_U.MPropMax: " << +TRPC_Long_Itg_U.MPropMax<< std::endl;
}




uint8_T TRPC_Long_Run_Counter = 0;

const uint64_t TRPCLong_LogFileDuration = 240000; // seconds

int main() {
    dpc::VED4DPC trpclong_data_from_VED;
    dpc::Decision4Long  trpclong_data_from_DM;
    dpc::DM2LongRefLine dm_2_long_ref_line;
    dpc::TargetsInfo trpclong_data_from_EM_TargetsInfo;
    dpc::TargetInfo trpclong_data_from_EM_TargetInfo;
    dpc::RoadEdgeInfo trpclong_data_from_Em_RoadEdgeInfo;
    dpc::LaneMarkerInfo trpclong_data_from_EM_LaneMarkerInfo;
    CAN2SWC CAN_data_4_TRPCLong;

    PendingSignals TargerSpeed_enum = IPC_Target_Speed;
    SWC2CAN TargerSpeed;

    PendingSignals TargerSpeedValid_enum = IPC_Target_SpdValid;
    SWC2CAN TargerSpeedValid;

    PendingSignals TargetTorque_enum = IPC_Target_Torque;
    SWC2CAN TargetTorque;

    PendingSignals BrakePressure_enum = IPC_Target_Brk_Press;
    SWC2CAN BrakePressure;

    PendingSignals ControlMode_enum = IPC_Target_Mode;
    SWC2CAN ControlMode;

    PendingSignals TargetGear_enum = IPC_Target_Gear;
    SWC2CAN TargetGear;

    PendingSignals DriveMotorEnable_enum = IPC_Target_Motor_En;
    SWC2CAN DriveMotorEnable;

    PendingSignals SpeedMode_enum = IPC_Target_SpdMode;
    SWC2CAN SpeedMode;

    PendingSignals LifeCycle_enum = IPC_Life_Signal;
    SWC2CAN LifeCycle;
    uint8_t life_cycle = 0;

    PendingSignals TorqueOrRPMMode_enum = IPC_Torque_Rpm_Mode;
    SWC2CAN TorqueOrRPMMode;

    PendingSignals TurnLight_enum = IPC_TurnLight;
    SWC2CAN TurnLight;
    
    TRPCLongSWC2CANStruct LongToCANstruct4log;
    
    std::ofstream TRPCLongLogFile;
    TRPCLongLogFile.open(LogFileNameDated("TRPCLongLogFile"), std::ios::out);
    uint8_t flag_First = 1;

    uint32_t TRPCLong_run_counter = 0;

    uint64_t TRPCLong_dataLoggingTimerStart = getCurrentTimeStamp();

    ara::log::InitLogging("DPC", "EM_TEST", ara::log::LogLevel::kVerbose, (ara::log::LogMode::kConsole | ara::log::LogMode::kRemote));

	ara::log::Logger &mainLog { ara::log::CreateLogger("main", "em sample context", ara::log::LogLevel::kVerbose) };

	mainLog.LogInfo() << "Let's produce some trpc long data...";

	GeneralWrapper gw(mainLog);

	const double torque_saturation_min = 0.0;
	const double torque_saturation_max = 12.5;
	PID_Controller PID_torque(1.3,0.0,0.01);

    while (1)
    {
        if (TRPCLong_run_counter == 0) {
            TRPC_Long_Itg_initialize();
        }
        else {
            gw._VED2TRPCLong_Required.s.GetData(trpclong_data_from_VED);
            gw._DM2TRPCLong_Required.s.GetData(trpclong_data_from_DM);
            gw._DM2LongRefLine_Required.s.GetData(dm_2_long_ref_line);
            gw._EM2TRPCLongTargetsInfo_Required.s.GetData(trpclong_data_from_EM_TargetsInfo);
            gw._EM2TRPCLongTargetInfo_Required.s.GetData(trpclong_data_from_EM_TargetInfo);
            gw._EM2TRPCLongRoadEdgeInfo_Required.s.GetData(trpclong_data_from_Em_RoadEdgeInfo);
            gw._EM2TRPCLongLaneMarkerInfo_Required.s.GetData(trpclong_data_from_EM_LaneMarkerInfo);
            gw._CAN2Long_Required.s.GetData(CAN_data_4_TRPCLong);
            
            // TODO: Control data sync --> Must have the same arival time
            assignTRPCLongInputsFromCAN(CAN_data_4_TRPCLong);
            assignTRPCLongInputsFromVED(trpclong_data_from_VED);
            assignTRPCLongInputsFromBAandEM(trpclong_data_from_DM, 
                                            trpclong_data_from_EM_TargetsInfo, 
                                            trpclong_data_from_EM_TargetInfo, 
                                            trpclong_data_from_Em_RoadEdgeInfo, 
                                            trpclong_data_from_EM_LaneMarkerInfo);

            TRPC_Long_Itg_step();
        }

        // TODO publish data

        ControlMode.signal = (UInt8)(ControlMode_enum);
        uint8_t control_mode = 1;
        // memcpy(&(ControlMode.value), &(control_mode), 
        //                         sizeof(ControlMode.value));
        ControlMode.value = static_cast<uint64_t>(control_mode);
        gw._TRPCLong2CAN_Provide.s.SetData(ControlMode);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.ControlMode = ControlMode;
        LongToCANstruct4log.control_mode = control_mode;

        TargetGear.signal = (UInt8)(TargetGear_enum);
        DriveMotorEnable.signal = (UInt8)(DriveMotorEnable_enum);
        uint8_t target_gear_value = 1;
        uint8_t drive_motor_enable = 0;
        if (/*TRPC_Long_Itg_Y.AccVlcMainReq*/ TRPC_Long_Itg_Y.VlcECEnable == 1) {
            target_gear_value = 1;
            drive_motor_enable = 1;
        }
        // memcpy(&(TargetGear.value), &(target_gear_value), 
        //                         sizeof(TargetGear.value));
        TargetGear.value = static_cast<uint64_t>(target_gear_value);
        // memcpy(&(DriveMotorEnable.value), &(drive_motor_enable), 
        //                         sizeof(DriveMotorEnable.value));
        DriveMotorEnable.value = static_cast<uint64_t>(drive_motor_enable);
        gw._TRPCLong2CAN_Provide.s.SetData(TargetGear);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.TargetGear = TargetGear;
        LongToCANstruct4log.target_gear_value = target_gear_value;
        gw._TRPCLong2CAN_Provide.s.SetData(DriveMotorEnable);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.DriveMotorEnable = DriveMotorEnable;
        LongToCANstruct4log.drive_motor_enable = drive_motor_enable;

        TorqueOrRPMMode.signal = (UInt8)(TorqueOrRPMMode_enum);
        uint8_t torque_or_rpm_mode = 0;
        // memcpy(&(TorqueOrRPMMode.value), &(torque_or_rpm_mode), 
        //                         sizeof(TorqueOrRPMMode.value));
        TorqueOrRPMMode.value = static_cast<uint64_t>(torque_or_rpm_mode);
        gw._TRPCLong2CAN_Provide.s.SetData(TorqueOrRPMMode);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.TorqueOrRPMMode = TorqueOrRPMMode;
        LongToCANstruct4log.torque_or_rpm_mode = torque_or_rpm_mode;
        
        SpeedMode.signal = (UInt8)(SpeedMode_enum);
        uint8_t speed_mode = 0;

        std::cout << "TRPC_Long_Itg_Y.Vx_des_tar : " << TRPC_Long_Itg_Y.Vx_des_tar << std::endl;
        // TRPC_Long_Itg_Y.Vx_des_tar = 5;
//

        if (TRPC_Long_Itg_Y.Vx_des_tar >= 0. && TRPC_Long_Itg_Y.Vx_des_tar < 5.) {speed_mode = 0;}
        else if (TRPC_Long_Itg_Y.Vx_des_tar >= 5. && TRPC_Long_Itg_Y.Vx_des_tar < 10.) {speed_mode = 1;}
        else {speed_mode = 2;}
        // memcpy(&(SpeedMode.value), &(speed_mode), 
        //                         sizeof(SpeedMode.value));
        SpeedMode.value = static_cast<uint64_t>(speed_mode);
        gw._TRPCLong2CAN_Provide.s.SetData(SpeedMode);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.SpeedMode = SpeedMode;
        LongToCANstruct4log.speed_mode = speed_mode;

        TargerSpeed.signal = (UInt8)(TargerSpeed_enum);
        double Vx_des_tar_double = static_cast<double>(TRPC_Long_Itg_Y.Vx_des_tar);
        memcpy(&(TargerSpeed.value), &(Vx_des_tar_double), 
                                sizeof(TargerSpeed.value));
        std::cout << "target speed value2: " << Vx_des_tar_double << std::endl;
        gw._TRPCLong2CAN_Provide.s.SetData(TargerSpeed);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.TargerSpeed = TargerSpeed;
        LongToCANstruct4log.Vx_des_tar = TRPC_Long_Itg_Y.Vx_des_tar;


        double target_torque_PID;
        target_torque_PID = PID_torque.compute(trpclong_data_from_VED.LongitudinalVelocity, Vx_des_tar_double, 0.01);
        if (target_torque_PID > torque_saturation_max) {target_torque_PID = torque_saturation_max;}
        else if(target_torque_PID < torque_saturation_min) {target_torque_PID = torque_saturation_min;}
        std::cout << "QQQQ Computed Torque by PID: " << target_torque_PID << "\n";


        TargerSpeedValid.signal = (UInt8)(TargerSpeedValid_enum);
        uint8_t TargerSpeedValid_value = 0; // Faked it as we use torque only
        // memcpy(&(TargerSpeedValid.value), &(TargerSpeedValid_value), 
        //                         sizeof(TargerSpeedValid.value));
        TargerSpeedValid.value = static_cast<uint64_t>(TargerSpeedValid_value);
        gw._TRPCLong2CAN_Provide.s.SetData(TargerSpeedValid);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.TargerSpeedValid = TargerSpeedValid;
        LongToCANstruct4log.TargerSpeedValid_value = TargerSpeedValid_value;

        // double MProp_saturated;
        // MProp_saturated = TRPC_Long_Itg_Y.MProp;
        // if (MProp_saturated > torque_saturation_max) {MProp_saturated = torque_saturation_max;}
        // else if(MProp_saturated < torque_saturation_min) {MProp_saturated = torque_saturation_min;}
        TargetTorque.signal = (UInt8)(TargetTorque_enum);
        memcpy(&(TargetTorque.value), &(/*MProp_saturated/*target_torque_PID*/TRPC_Long_Itg_Y.MProp),
                                sizeof(TargetTorque.value));
        gw._TRPCLong2CAN_Provide.s.SetData(TargetTorque);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.TargetTorque = TargetTorque;
        LongToCANstruct4log.MProp = TRPC_Long_Itg_Y.MProp; // target_torque_PID; //TRPC_Long_Itg_Y.MProp;
        std::cout << "QQQQ MProp Torque by Long: " << TRPC_Long_Itg_Y.MProp << "\n";
        // std::cout << "QQQQ MProp Torque by Long saturated: " << MProp_saturated << "\n";
        std::cout << "QQQQ DC Enable by Long: " << +TRPC_Long_Itg_Y.VlcDCEnable << "\n";
        std::cout << "QQQQ EC Enable by Long: " << +TRPC_Long_Itg_Y.VlcECEnable << "\n";


        BrakePressure.signal = (UInt8)(BrakePressure_enum);
        TurnLight.signal = (UInt8)(TurnLight_enum);
        uint8_t turn_signal = 0;
        double param_decel_2_pressure = 10.;  // Can be tuned later!
        double brake_pressure_command = 0.;
        if (TRPC_Long_Itg_Y.VlcDCEnable == 1 /*&& TRPC_Long_Itg_Y.Ax_tar < 0*/) {
            // brake_pressure_command = static_cast<double>(TRPC_Long_Itg_Y.Ax_tar) * (-0.24) + 0.4;
            brake_pressure_command = static_cast<double>(TRPC_Long_Itg_Y.VLC_a_TRP2VLC_Ax_Tar_j);
            std::cout << "QQQQ Ax_tar by Long: " << TRPC_Long_Itg_Y.Ax_tar << "\n";
            std::cout << "QQQQ Original brake_pressure_command by Long (SWC2CAN): " << brake_pressure_command << "\n";
            if (brake_pressure_command < 0) {brake_pressure_command = 0;} 
            std::cout << "QQQQ Mapped brake_pressure_command by Long (SWC2CAN): " << brake_pressure_command << "\n";
            turn_signal = 0;
        }
        memcpy(&(BrakePressure.value), &(brake_pressure_command), 
                                sizeof(BrakePressure.value));
        gw._TRPCLong2CAN_Provide.s.SetData(BrakePressure);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.BrakePressure = BrakePressure;
        LongToCANstruct4log.brake_pressure_command = brake_pressure_command;


        // memcpy(&(TurnLight.value), &(turn_signal), 
        //                     sizeof(TurnLight.value));
        TurnLight.value = static_cast<uint64_t>(turn_signal);
        gw._TRPCLong2CAN_Provide.s.SetData(TurnLight);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.TurnLight = TurnLight;
        LongToCANstruct4log.turn_signal = turn_signal;


        LifeCycle.signal = (UInt8)(LifeCycle_enum);
        // memcpy(&(LifeCycle.value), &(life_cycle), 
        //                         sizeof(LifeCycle.value));
        LifeCycle.value = static_cast<uint64_t>(life_cycle);
        gw._TRPCLong2CAN_Provide.s.SetData(LifeCycle);
        gw._TRPCLong2CAN_Provide.Publish();
        LongToCANstruct4log.LifeCycle = LifeCycle;
        LongToCANstruct4log.life_cycle = life_cycle;
        life_cycle++;
        if(life_cycle>15) {
            life_cycle = 0; 
        }              


        TRPCLong_run_counter++;

        uint64_t TRPCLong_dataLoggingTimerNow = getCurrentTimeStamp();
        if ((TRPCLong_dataLoggingTimerNow - TRPCLong_dataLoggingTimerStart)<= TRPCLong_LogFileDuration) {
            logTRPCLongData(&TRPCLongLogFile,
                            flag_First, 
                            &trpclong_data_from_VED, 
                            &trpclong_data_from_DM,
                            &dm_2_long_ref_line, 
                            &trpclong_data_from_EM_TargetsInfo, 
                            &trpclong_data_from_EM_TargetInfo,
                            &trpclong_data_from_Em_RoadEdgeInfo, 
                            &trpclong_data_from_EM_LaneMarkerInfo,
                            &LongToCANstruct4log);
        }
        else {
            TRPCLongLogFile.close();
        }
        // TODO, check the trpclong cycle time
        flag_First = 0;
        std::this_thread::sleep_for(std::chrono::milliseconds(10)); // TRPCLong cycle time: ? ms
    }
    

    return 0;
}

std::string LogFileNameDated(const std::string& inputName) {
    time_t now = time(0);
    tm* localTime = localtime(&now);

    int year = 1900 + localTime->tm_year;
    int month = 1 + localTime->tm_mon;
    int day = localTime->tm_mday;
    int hour = localTime->tm_hour;
    int min = localTime->tm_min;
    int sec = localTime->tm_sec;

    std::string fileName = inputName + "_" + std::to_string(year) + "_" +
                           std::to_string(month) + "_" + std::to_string(day) + "_" +
                           std::to_string(hour) + "_" + std::to_string(min) + "_" +
                           std::to_string(sec) + ".csv";

    return fileName;
}

// TODO log data
void logTRPCLongData(std::ofstream *ptr_File, uint8_t flag_First, dpc::VED4DPC* const ptr_trpclong_data_from_VED,
                    dpc::Decision4Long* const  ptr_trpclong_data_from_DM,
                    dpc::DM2LongRefLine* const  trpclong_data_from_DM_refLine,
                    dpc::TargetsInfo* const ptr_trpclong_data_from_EM_TargetsInfo, 
                    dpc::TargetInfo* const ptr_trpclong_data_from_EM_TargetInfo, 
                    dpc::RoadEdgeInfo* const ptr_trpclong_data_from_Em_RoadEdgeInfo, 
                    dpc::LaneMarkerInfo* const ptr_trpclong_data_from_EM_LaneMarkerInf,
                    TRPCLongSWC2CANStruct* const ptr_trpclong_data_to_CAN
					){

	if (flag_First == 1) {
					*ptr_File << "Timestamp" << ",";

					//Log main received data From VED
					// *ptr_File << ptr_trpclong_data_from_VED->  // todo, add ivehsdslsts to log
					*ptr_File << "LongitudinalVelocity" << ",";
					*ptr_File << "LongitudinalAcceleration" << ",";
					*ptr_File << "LateralVelocity" << ","; //here we lat and lon acc seperated, do we need the overal acc?
					*ptr_File << "LateralAcceleration" << ",";
					*ptr_File << "MotionState" << ",";
					*ptr_File << "SteeringAngle" << ",";
					*ptr_File << "SteeringRate" << ",";
					*ptr_File << "Yawrate" << ",";

					//Log main received data from DM
					*ptr_File << "ba_trp_v_sv_setspd" << ",";
					*ptr_File << "ba_trp_indx_trp_lgt_sts" << ",";
					*ptr_File << "ba_trp_a_tgt_longt_accel_iv" << ",";
					*ptr_File << "ba_trp_indx_tgt_id_iv" << ",";
					*ptr_File << "ba_trp_v_tgt_longt_spd_iv" << ",";
					*ptr_File << "ba_trp_d_tgt_longt_dist_iv" << ",";
					*ptr_File << "ba_trp_indx_spdlim_typ" << ",";
					*ptr_File << "ba_trp_kmh_spdlim_act_val" << ",";
					*ptr_File << "ba_trp_d_spdlim_nxt_dst" << ",";
					*ptr_File << "ba_trp_indx_opt_action" << ",";
                    *ptr_File << "ba_ref_line_c0" << ",";
                    *ptr_File << "ba_ref_line_c1" << ",";
                    *ptr_File << "ba_ref_line_c2" << ",";
                    *ptr_File << "ba_ref_line_c3" << ",";

					//Log main received data from EM
					*ptr_File << "indx_tgtid" << ",";
					*ptr_File << "d_tgtlongtdist" << ",";
					*ptr_File << "v_tgtlongtspd" << ",";
					*ptr_File << "a_tgtlongtaccel" << ",";

					//Log the Can Command from TRPCLon
					*ptr_File << "ControlMode.signal"<< ",";
					*ptr_File << "ControlMode.value"<< ",";
					*ptr_File << "control_mode" << ",";

					*ptr_File << "TargetGear.signal" << ",";
					*ptr_File << "TargetGear.value" << ",";
					*ptr_File << "target_gear_value" << ",";

					*ptr_File << "DriveMotorEnable.signal"<< ",";
					*ptr_File << "DriveMotorEnable.value"<< ",";
					*ptr_File << "drive_motor_enable"<< ",";

					*ptr_File << "TorqueOrRPMMode.signal"<< ",";
					*ptr_File << "TorqueOrRPMMode.value"<< ",";
					*ptr_File << "torque_or_rpm_mode"<< ",";

					*ptr_File << "SpeedMode.signal"<< ",";
					*ptr_File << "SpeedMode.value"<< ",";
					*ptr_File << "speed_mode"<< ",";

					*ptr_File << "TargerSpeed.signal"<< ",";
					*ptr_File << "TargerSpeed.value"<< ",";
					*ptr_File << "Vx_des_tar"<< ",";

					*ptr_File << "TargerSpeedValid.signal"<< ",";
					*ptr_File << "TargerSpeedValid.value"<< ",";
					*ptr_File << "TargerSpeedValid_value"<< ",";

					*ptr_File << "TargetTorque.signal"<< ",";
					*ptr_File << "TargetTorque.value"<< ",";
					*ptr_File << "MProp"<< ",";

					*ptr_File << "BrakePressure.signal"<< ",";
					*ptr_File << "BrakePressure.value"<< ",";
					*ptr_File << "brake_pressure_command"<< ",";

					*ptr_File << "TurnLight.signal"<< ",";
					*ptr_File << "TurnLight.value"<< ",";
					*ptr_File << "turn_signal"<< ",";

					*ptr_File << "LifeCycle.signal"<< ",";
					*ptr_File << "LifeCycle.value"<< ",";
					*ptr_File << "life_cycle";

					//Go to the next line
					*ptr_File << "\n";
	}
	else{

                        *ptr_File << +getCurrentTimeStamp() << ",";

                        //Log main received data From VED
                        // *ptr_File << ptr_trpclong_data_from_VED->  // todo, add ivehsdslsts to log
                        *ptr_File << +ptr_trpclong_data_from_VED->LongitudinalVelocity << ",";
                        *ptr_File << +ptr_trpclong_data_from_VED->LongitudinalAcceleration << ",";
                        *ptr_File << +ptr_trpclong_data_from_VED->LateralVelocity << ","; //here we lat and lon acc seperated, do we need the overal acc?
                        *ptr_File << +ptr_trpclong_data_from_VED->LateralAcceleration << ",";
                        *ptr_File << +ptr_trpclong_data_from_VED->MotionState << ",";
                        *ptr_File << +ptr_trpclong_data_from_VED->SteeringAngle << ",";
                        *ptr_File << +ptr_trpclong_data_from_VED->SteeringRate << ",";
                        *ptr_File << +ptr_trpclong_data_from_VED->Yawrate << ",";
                        
                        //Log main received data from DM
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_v_sv_setspd << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_indx_trp_lgt_sts << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_a_tgt_longt_accel_iv << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_indx_tgt_id_iv << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_v_tgt_longt_spd_iv << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_d_tgt_longt_dist_iv << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_indx_spdlim_typ << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_kmh_spdlim_act_val << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_d_spdlim_nxt_dst << ",";
                        *ptr_File << +ptr_trpclong_data_from_DM->ba_trp_indx_opt_action << ",";
                        *ptr_File << +trpclong_data_from_DM_refLine->ba_ref_line_c0 << ",";
                        *ptr_File << +trpclong_data_from_DM_refLine->ba_ref_line_c1 << ",";
                        *ptr_File << +trpclong_data_from_DM_refLine->ba_ref_line_c2 << ",";
                        *ptr_File << +trpclong_data_from_DM_refLine->ba_ref_line_c3 << ",";

                        //Log main received data from EM
                        *ptr_File << +ptr_trpclong_data_from_EM_TargetInfo->indx_tgtid << ",";
                        *ptr_File << +ptr_trpclong_data_from_EM_TargetInfo->d_tgtlongtdist << ",";
                        *ptr_File << +ptr_trpclong_data_from_EM_TargetInfo->v_tgtlongtspd << ",";
                        *ptr_File << +ptr_trpclong_data_from_EM_TargetInfo->a_tgtlongtaccel << ",";

                        //Log the Can Command from TRPCLon 
                        *ptr_File << +ptr_trpclong_data_to_CAN->ControlMode.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->ControlMode.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->control_mode<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->TargetGear.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->TargetGear.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->target_gear_value<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->DriveMotorEnable.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->DriveMotorEnable.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->drive_motor_enable<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->TorqueOrRPMMode.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->TorqueOrRPMMode.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->torque_or_rpm_mode<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->SpeedMode.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->SpeedMode.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->speed_mode<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->TargerSpeed.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->TargerSpeed.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->Vx_des_tar<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->TargerSpeedValid.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->TargerSpeedValid.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->TargerSpeedValid_value<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->TargetTorque.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->TargetTorque.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->MProp<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->BrakePressure.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->BrakePressure.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->brake_pressure_command<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->TurnLight.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->TurnLight.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->turn_signal<< ",";

                        *ptr_File << +ptr_trpclong_data_to_CAN->LifeCycle.signal<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->LifeCycle.value<< ",";
                        *ptr_File << +ptr_trpclong_data_to_CAN->life_cycle<< ",";

                        //Go to the next line
                        *ptr_File << "\n";
                    }
}

uint64_t getCurrentTimeStamp() {
//    auto current_time = std::chrono::system_clock::now();
//    auto timestamp = std::chrono::duration_cast<std::chrono::seconds>(current_time.time_since_epoch());
//    uint32_t stamp_sec = static_cast<uint32_t>(timestamp.count());
//    uint32_t stamp_nanosec = static_cast<uint32_t>((current_time.time_since_epoch() - timestamp).count());
//    double CurrentTimeStamp = stamp_sec + (stamp_nanosec * 1e-9);
//    return CurrentTimeStamp;
	std::chrono::high_resolution_clock m_clock;
	return std::chrono::duration_cast<std::chrono::milliseconds>(m_clock.now().time_since_epoch()).count();
    }
